from django import template
from django.utils.safestring import mark_safe

from process.conf import get_conf
from ..models import Process, Job, JobTask

register = template.Library()

# noinspection SpellCheckingInspection
html = """
<div id="container-process-{id}"></div>
<script>
Highcharts.chart('container-process-{id}', {
  chart: {
    height: """ + get_conf('diagram__chart_height') + """,
    inverted: true
  },
  title: {
    useHTML: true,
    text: '{name}'
  },
  series: [{
    type: 'organization',
    name: '{name}',
    keys: ['from', 'to'],
    data: {data},
    levels: [],
    linkRadius: """ + get_conf('diagram__link_radius') + """,
    linkLineWidth: """ + get_conf('diagram__link_line_width') + """,
    linkColor: '""" + get_conf('diagram__link_color') + """',
    nodes: {nodes},
    showCheckbox: true,
    colorByPoint: false,
    color: '#007ad0',
    dataLabels: {
      color: 'white',
    },
    borderColor: 'white',
    nodeWidth: """ + get_conf('diagram__node_width') + """,
    nodePadding: """ + get_conf('diagram__node_padding') + """
  }],
  tooltip: {
    outside: true,
    formatter: function() {
      //debugger;
      return this.point.info;
    }
  },
  exporting: {
    allowHTML: true,
    sourceWidth: 800,
    sourceHeight: 600
  }
});
</script>
"""


def get_task_as_node(t):
    job_task = True if isinstance(t, JobTask) else False
    node_task = t.task if job_task else t
    color = get_conf(f'diagram__tasks_color__{t.status}') if job_task else get_conf('diagram__tasks_color__default')
    return {
        'id': node_task.name,
        'name': node_task.name,
        'title': t.title if job_task else node_task.description,
        'level': node_task.level,
        'offset': node_task.offset,
        'info': t.info if job_task else node_task.description,
        'color': color,
    }


@register.filter(name='diagram', is_safe=True)
def diagram(obj):
    try:
        assert isinstance(obj, Process), "err"
        process = obj
    except AssertionError:
        assert isinstance(obj, Job), "argument is not Process object either Job object"
        process = obj.process

    data = []
    nodes = []
    # noinspection PyUnresolvedReferences
    for task in obj.tasks.all():
        nodes.append(get_task_as_node(task))
        tk = task if isinstance(obj, Process) else task.task
        if tk.childs.all().count():
            for child in tk.childs.all():
                data.append([str(tk.name), str(child.task.name)])
        else:
            data.append([str(tk.name), str(tk.name)])

    data.sort()
    response = html.replace('{id}', str(process.id))
    response = response.replace('{name}', str(obj.__str__()))
    response = response.replace('{data}', str(data))
    response = response.replace('{nodes}', str(nodes))
    return mark_safe(response)



import os
import sys
import logging
import subprocess
from threading import Thread
from django.utils import timezone

from process.conf import get_conf
from process.models import Job, JobTask
logger = logging.getLogger('django-process')


class TaskThreaded(Thread):
    def __init__(self, obj, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.obj = obj

    def run(self):
        try:
            try:
                # get interpreter the default is the one used to run django
                if not self.obj.task.interpreter:
                    cmd = [sys.executable]
                else:
                    cmd = self.obj.task.interpreter.split()

                # noinspection SpellCheckingInspection
                if self.obj.task.code.file.__class__.__name__ == 'S3Boto3StorageFile':
                    if not os.path.isdir('/tmp/dj_process_tasks'):
                        os.makedirs('/tmp/dj_process_tasks')
                    file_path = os.path.join('/tmp', self.obj.task.code.file.name)
                    if not os.path.isfile(file_path):
                        with open(file_path, 'wb') as code:
                            code.write(self.obj.task.code.file.read())
                else:
                    file_path = self.obj.task.code.path

                # append task file path and arguments if they exists
                cmd.append(file_path)
                if self.obj.task.arguments:
                    cmd += self.obj.task.arguments.split()

                logger.info(f'command to execute {cmd}')

                p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                stdout, stderr = p.communicate()
                # return code must be 0 for success
                self.obj.observations = stdout.decode('utf-8')
                if p.returncode:
                    raise Exception(stderr.decode('utf-8'))

                self.obj.set_status(JobTask.finished)

            except Exception as e:
                # if error then send to logger and also mark task and it's job as error
                self.obj.observations += f"\nexception when running task {e}"
                self.obj.set_status(JobTask.error)
                self.obj.job.status = Job.error
                logger.error(f'task {self.obj} finished with error {self.obj.observations}')
                self.obj.job.save()
                # if there is a custom handler send task id and exception to it
                error_handler_func = get_conf('task__error_handler')
                logger.info(f'sending info to handler {error_handler_func.__name__}')
                error_handler_func(self.obj, e)

            self.obj.dt_end = timezone.now()
            self.obj.save()
        except Exception as e:
            logger.exception(f'error {e} when processing task {self.obj}')



import logging
from django.contrib import messages
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _
from django.views import View

from process.conf import get_conf
from .generic_views import (
    ProcessSecurity,
    ProcessGenericListView,
    ProcessGenericDeleteView
)
from process.models import Job

logger = logging.getLogger('django-process')


class JobListView(ProcessGenericListView):
    model = Job
    title = 'Jobs'
    filters = get_conf('views__job__list__url_allow_filters')
    permissions = get_conf('views__job__list__permissions')

    def post(self, request, *args, **kwargs):
        request = JobCancelView.as_view()(request)
        # noinspection PyTypeChecker
        return self.get(request, *args, **kwargs)


class JobDeleteView(ProcessGenericDeleteView):
    model = Job
    success_url = get_conf('views__job__delete__success_url')
    success_message = get_conf('views__job__delete__success_message')
    permissions = get_conf('views__job__delete__permissions')


class JobCancelView(ProcessSecurity, View):
    permissions = get_conf('views__job__cancel__permissions')

    # noinspection PyUnusedLocal
    def post(self, request, *args, **kwargs):
        logger.debug(f'cancel job i got post request=> {self.request.POST}')
        try:
            job = get_object_or_404(Job, id=self.request.POST['job'])
        except KeyError:
            messages.error(request, _('job does not exists'))
            return request

        try:
            job.cancel()
            messages.success(request, get_conf('views__job__cancel__success_message'))
        except Exception as e:
            logger.exception(f"can't cancel job due to error=> {e}")
            messages.error(request, _(f'{e}'))

        return request



from django.contrib import messages
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _
from django.views import View

from process.conf import get_conf
from .generic_views import (
    ProcessSecurity,
    ProcessGenericCreateView,
    ProcessGenericListView,
    ProcessGenericUpdateView,
    ProcessGenericDeleteView
)


from process.models import Process, Job


class ProcessCreateView(ProcessGenericCreateView):
    model = Process
    success_url = get_conf('views__process__create__success_url')
    success_message = get_conf('views__process__create__success_message')
    permissions = get_conf('views__process__create__permissions')


class ProcessListView(ProcessGenericListView):
    model = Process
    title = 'Processes'
    filters = get_conf('views__process__list__url_allow_filters')
    permissions = get_conf('views__process__list__permissions')

    def post(self, request, *args, **kwargs):
        request = ProcessRunOnDemandView.as_view()(request)
        # noinspection PyTypeChecker
        return self.get(request, *args, **kwargs)


class ProcessUpdateView(ProcessGenericUpdateView):
    model = Process
    success_url = get_conf('views__process__update__success_url')
    success_message = get_conf('views__process__update__success_message')
    permissions = get_conf('views__process__update__permissions')


class ProcessDeleteView(ProcessGenericDeleteView):
    model = Process
    success_url = get_conf('views__process__delete__success_url')
    success_message = get_conf('views__process__delete__success_message')
    permissions = get_conf('views__process__delete__permissions')


class ProcessRunOnDemandView(ProcessSecurity, View):
    permissions = get_conf('views__process__run__permissions')

    # noinspection PyUnusedLocal
    def post(self, request, *args, **kwargs):
        try:
            process = get_object_or_404(Process, id=self.request.POST['process'])
            __, __ = Job.create(process)
        except Exception as e:
            messages.error(request, _(f'{e}'))
        finally:
            return request



from django.conf.urls import url
from process.models import Job, Process
from .views import process, task, job, jobtask, diagram

urlpatterns = [
    # Processes
    url(r'^processes/$', process.ProcessListView.as_view(), name='process-processes'),
    url(r'^processes/add/$', process.ProcessCreateView.as_view(), name='process-processes-add'),
    url(r'^processes/(?P<pk>[0-9]+)/$', process.ProcessUpdateView.as_view(), name='process-processes-update'),
    url(r'^processes/(?P<pk>[0-9]+)/delete/$', process.ProcessDeleteView.as_view(), name='process-processes-delete'),
    url(r'^processes/(?P<pk>[0-9]+)/diagram/$', diagram.DiagramView.as_view(model=Process),
        name='process-process-diagram'),
    # Tasks
    url(r'^tasks/$', task.TaskListView.as_view(), name='process-tasks'),
    url(r'^tasks/add/$', task.TaskCreateView.as_view(), name='process-tasks-add'),
    url(r'^tasks/(?P<pk>[0-9]+)/$', task.TaskUpdateView.as_view(), name='process-tasks-update'),
    url(r'^tasks/(?P<pk>[0-9]+)/delete/$', task.TaskDeleteView.as_view(), name='process-tasks-delete'),
    # Jobs
    url(r'^jobs/$', job.JobListView.as_view(), name='process-jobs'),
    url(r'^jobs/(?P<pk>[0-9]+)/delete/$', job.JobDeleteView.as_view(), name='process-jobs-delete'),
    url(r'^jobs/(?P<pk>[0-9]+)/diagram/$', diagram.DiagramView.as_view(model=Job), name='process-job-diagram'),
    # JobTasks
    url(r'^job-tasks/$', jobtask.JobTaskListView.as_view(), name='process-job-tasks'),
]



import logging
from django.contrib import messages
from django.utils.decorators import method_decorator
from django.utils.translation import gettext_lazy as _
from django.shortcuts import get_object_or_404
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from process.conf import get_conf
from .generic_views import (
    ProcessSecurity,
    ProcessGenericListView,
    ProcessGenericDeleteView
)
from process.models import JobTask

logger = logging.getLogger('django-process')


# noinspection SpellCheckingInspection
class JobTaskDeleteView(ProcessGenericDeleteView):
    model = JobTask
    success_url = get_conf('views__jobtask__delete__success_url')
    success_message = get_conf('views__jobtask__delete__success_message')
    permissions = get_conf('views__jobtask__delete__permissions')


# noinspection SpellCheckingInspection
class JobTaskListView(ProcessGenericListView):
    model = JobTask
    title = 'JobTasks'
    filters = get_conf('views__jobtask__list__url_allow_filters')
    permissions = get_conf('views__jobtask__list__permissions')

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, args, kwargs)

    def post(self, request, *args, **kwargs):
        request = JobTaskManagementView.as_view()(request)
        # noinspection PyTypeChecker
        return self.get(request, *args, **kwargs)


# noinspection SpellCheckingInspection
class JobTaskManagementView(ProcessSecurity, View):
    permissions = get_conf('views__jobtask__management__permissions')

    # noinspection PyUnusedLocal
    def post(self, request, *args, **kwargs):
        try:
            logger.debug(f'cancel job i got post request=> {self.request.POST}')
            task = get_object_or_404(JobTask, id=self.request.POST['task'])
            action = self.request.POST['action']
            if action == JobTask.reopened:
                task.reopen(main=True)
            else:
                task.set_status(action)

            messages.success(request, get_conf('views__jobtask__management__success_message').format(action=action))
        except Exception as e:
            messages.error(request, _(f'{e}'))
        finally:
            return request



import logging
from django.contrib import messages
from django.urls import reverse_lazy
from django.views.generic.list import ListView
from django.utils.translation import gettext_lazy as _
from django.contrib.messages.views import SuccessMessageMixin
from django.contrib.auth.mixins import LoginRequiredMixin, UserPassesTestMixin
from django.views.generic.edit import UpdateView, CreateView, DeleteView

from process.conf import get_conf

logger = logging.getLogger('django-process')


# noinspection PyUnresolvedReferences
class ProcessSecurity(LoginRequiredMixin, UserPassesTestMixin):
    raise_exception = get_conf('views__security_raise_exception')

    def test_func(self):
        if hasattr(self, 'permissions'):
            usr = self.request.user
            [logger.debug(f'user {usr.username} permission {i} result {usr.has_perm(i)}') for i in self.permissions]
            return not any([not self.request.user.has_perm(i) for i in self.permissions])
        return True


class ProcessGenericListView(ProcessSecurity, ListView):
    title = ''
    permissions = []
    paginate_by = get_conf('views__paginate')
    template_name = get_conf('views__templates__objects_list')
    filters = {}

    def get_filters_from_request(self):
        # convert set to python dict
        request_dict = {k: v for k, v in self.request.GET.lists()}
        # search filters in url if filters do exists change key add __in
        return {f'{k}__in': request_dict.get(k) for k in self.filters if request_dict.get(k)}

    def get_queryset(self):
        filters = self.get_filters_from_request()

        if filters:
            return self.model.objects.filter(**filters)
        else:
            return self.model.objects.all()

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        context['title'] = self.title
        return context


# noinspection PyUnresolvedReferences
class ProcessGenericEditView(ProcessSecurity, SuccessMessageMixin):
    fields = '__all__'
    template_name = get_conf('views__templates__object_edit')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.success_url = reverse_lazy(self.success_url)

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['operation'] = self.operation
        return context


class ProcessGenericUpdateView(ProcessGenericEditView, UpdateView):
    operation = _('Update')


class ProcessGenericCreateView(ProcessGenericEditView, CreateView):
    operation = _('Create')


class ProcessGenericDeleteView(ProcessSecurity, SuccessMessageMixin, DeleteView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.template_name = get_conf('views__templates__object_delete')
        self.success_url = reverse_lazy(self.success_url)

    def delete(self, request, *args, **kwargs):
        if hasattr(self, 'success_message'):
            messages.success(self.request, self.success_message)
        return super().delete(request, *args, **kwargs)



import time
import logging
from django.core.management.base import BaseCommand
from django.utils import timezone

from ._actions import configure_env, run_jobs, run_awaiting_tasks, finish_jobs

logger = logging.getLogger('django-process')


class Command(BaseCommand):
    help = 'Run All Jobs'

    def handle(self, *args, **options):
        logger.info('django-process run_jobs started')
        configure_env()
        try:
            # This var will prevent many jobs started from the same process in the 0 second
            jobs_start_unlocked = False
            while True:
                time.sleep(0.1)

                now = timezone.now()
                if now.second == 0 and jobs_start_unlocked:
                    jobs_start_unlocked = False
                    run_jobs(now)
                elif now.second != 0:
                    jobs_start_unlocked = True

                run_awaiting_tasks()
                finish_jobs()

        except KeyboardInterrupt:
            pass



import os
import sys
import json
import django
import logging

logger = logging.getLogger('django-process')

location = os.path.dirname(os.path.abspath(__file__))
file = os.path.join(location, 'env_conf.json')

if not os.path.isfile(file):
    logger.warning('env for django-process is not configured')
else:
    with open(file, 'r') as f:
        environment = json.load(f)

    if environment['project_path'] not in sys.path:
        sys.path.insert(0, environment['project_path'])

    os.environ.setdefault("DJANGO_SETTINGS_MODULE", environment['project_settings'])

if os.environ.get('DJANGO_SETTINGS_MODULE'):
    django.setup()



import logging
from django.conf import settings
from django.utils.translation import gettext_lazy as _

"""
when the module is loaded it will search in django's settings the dictionary DJANGO_PROCESS = {}
if it founds it, then it will override defaults values so the get_conf function will read from
defaults values.
"""

loaded = False

logger = logging.getLogger('django-process')


def override(defaults, custom):
    try:
        for k, v in custom.items():
            if isinstance(v, dict):
                override(defaults[k], v)
            else:
                defaults[k] = v
    except KeyError as e:
        e.args = (f'The key you were trying to override "{e.args[0]}" does not exists in defaults',)
        raise


def get_conf(conf):
    conf_tree = []
    try:
        conf_tree = conf.split('__')
        conf_src = default_settings.copy()
        for conf in conf_tree:
            conf_src = conf_src[conf]
        return conf_src
    except KeyError as e:
        e.args = (f"configuration does not exists {'/'.join(conf_tree)}",)
        raise


# noinspection PyUnusedLocal
def dummy_task_error_handler(task_id, exception): pass


# noinspection SpellCheckingInspection
default_settings = {
    'task': {
        'error_handler': dummy_task_error_handler
    },
    'views': {
        'paginate': 20,
        'security_raise_exception': True,
        'templates': {
            'objects_list': 'process/objects_list.html',
            'object_edit': 'process/object_edit.html',
            'object_delete': 'process/object_delete.html',
            'object_diagram': 'process/object_diagram.html',
            'task_edit': 'process/task_edit.html',
            'extension_images': {
                'pl': 'process/images/perl.png',
                'py': 'process/images/python.png',
                'sh': 'process/images/bash.png',
            },
        },
        'process': {
            'run': {
                'permissions': [
                    'process.view_processes',
                    'process.manage_processes',
                    'process.view_jobs',
                    'process.run_jobs'
                ],
            },
            'create': {
                'success_url': 'process-processes',
                'success_message': _('process created successfully'),
                'permissions': ['process.view_processes', 'process.add_process'],
            },
            'list': {
                'url_allow_filters': {'id': []},
                'permissions': ['process.view_processes'],
            },
            'update': {
                'success_url': 'process-processes',
                'success_message': _('process updated successfully'),
                'permissions': ['process.view_processes', 'process.change_process'],
            },
            'delete': {
                'success_url': 'process-processes',
                'success_message': _('process deleted successfully'),
                'permissions': ['process.view_processes', 'process.delete_process'],
            },
        },
        'task': {
            'create': {
                'success_url': 'process-tasks',
                'success_message': _('task created successfully'),
                'redirect_to_edit': True,
                'permissions': ['process.view_tasks', 'process.add_task'],
            },
            'list': {
                'url_allow_filters': {'id': [], 'process__id': []},
                'permissions': ['process.view_tasks'],
            },
            'update': {
                'success_url': 'process-tasks',
                'success_message': _('task updated successfully'),
                'permissions': ['process.view_tasks', 'process.change_task'],
            },
            'delete': {
                'success_url': 'process-tasks',
                'success_message': _('task deleted successfully'),
                'permissions': ['process.view_tasks', 'process.delete_task'],
            },
        },
        'job': {
            'list': {
                'url_allow_filters': {'id': [], 'process__id': [], 'status': []},
                'permissions': ['process.view_jobs'],
            },
            'delete': {
                'success_url': 'process-jobs',
                'success_message': _('job deleted successfully'),
                'permissions': ['process.view_jobs', 'process.delete_job'],
            },
            'cancel': {
                'success_message': _('job cancelled successfully'),
                'permissions': ['process.cancel_jobs'],
            },
        },
        'jobtask': {
            'list': {
                'url_allow_filters': {'id': [], 'job__id': [],  'task__id': [], 'status': []},
                'permissions': ['process.view_job_tasks'],
            },
            'delete': {
                'success_url': 'process-job-tasks',
                'success_message': _('task deleted successfully'),
                'permissions': ['process.view_job_tasks', 'process.delete_jobtask'],
            },
            'management': {
                'success_message': _('task successfully {action}'),
                'permissions': ['process.view_job_tasks', 'process.manage_job_tasks'],
            }
        },
    },
    'diagram': {
        'chart_height': '600',
        'node_width': '75',
        'node_padding': '0',
        'link_radius': '55',
        'link_line_width': '4',
        'link_color': 'black',
        'tasks_color': {
            'default': '#41c0a4',
            'initialized': '#419dc0',
            'awaiting': 'gray',
            'reopened': '#41c0a4',
            'retry': '#41c0a4',
            'finished': '#abd734',
            'cancelled': '#d76034',
            'forced': '#d734ab',
            'error': 'red',
        }
    }
}

if not loaded:
    try:
        logger.debug('overriding default settings of django-process')
        custom_settings = settings.DJANGO_PROCESS
        override(default_settings, custom_settings)
        loaded = True
    except AttributeError:
        pass



import re
import os
import logging
from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator, FileExtensionValidator
from django.db import models, transaction, DatabaseError
from django.utils import timezone
from django.utils.translation import gettext_lazy as _


from process.exceptions import ProcessException
logger = logging.getLogger('django-process')


# noinspection SpellCheckingInspection
class Process(models.Model):
    """
    The Process class this will define the execute options for the process
    """
    name = models.CharField(_("name"), max_length=20, unique=True)
    description = models.CharField(_("description"), max_length=200, unique=True)
    is_active = models.BooleanField(_("active"), default=True)
    run_if_err = models.BooleanField(_("run again if previously execution status is error"), default=False)
    run_overlap = models.BooleanField(_("run if previous still is executing"), default=False)
    minute = models.CharField(_("minute"), max_length=50)
    hour = models.CharField(_("hour"), max_length=50)
    day_of_month = models.CharField(_("day of month"), max_length=50)
    month = models.CharField(_("month"), max_length=50)
    day_of_week = models.CharField(_("day of week"), max_length=50)
    objects = models.Manager()

    def __str__(self):
        return self.name

    class Meta:
        db_table = 'pr_processes'
        verbose_name = _('process')
        verbose_name_plural = _('processes')
        ordering = ['-id']
        permissions = (
            ("view_processes", "Can view processes"),
            ("manage_processes", "Can manage processes"),
        )

    def clean(self):
        super().clean()
        _ = self._expanded(self.minute, 'minute')
        _ = self._expanded(self.hour, 'hour')
        _ = self._expanded(self.day_of_month, 'day_of_month')
        _ = self._expanded(self.month, 'month')
        _ = self._expanded(self.day_of_week, 'day_of_week')

    def save(self, *args, **kwargs):
        self.full_clean()
        return super().save(*args, **kwargs)

    @staticmethod
    def _expanded(var, type_val):
        """
        gets an crontab like string and expands all items example
        '1,2,3,7,4-9' will be expanded to '1,2,3,4,5,6,7,8,9'
        :param var:
        :return: string of unique items expanded
        """
        def item_validator(item):
            val_range = [0, 0]
            if type_val == 'minute':
                val_range = [0, 59]
            if type_val == 'hour':
                val_range = [0, 23]
            elif type_val == 'day_of_month':
                val_range = [1, 31]
            elif type_val == 'month':
                val_range = [1, 12]
            elif type_val == 'day_of_week':
                val_range = [0, 6]

            if not val_range[0] <= int(item) <= val_range[1]:
                raise ValueError
            return int(item)
            
        # if content it is a star it is ok means all
        if var == '*':
            return []

        n3 = []
        ranges = []
        # separate ranged values and not ranged values to be validated
        for i in var.split(','):
            if i.find('-') > -1 and len(re.findall('-', i)) == 1:
                ranges.append(i.split('-'))
            elif i.find('-') == -1:
                try:
                    n3.append(item_validator(i))
                except ValueError:
                    raise ValidationError(_(f'{i} is not a valid value'))
            else:
                raise ValidationError(_(f'{i} is not a valid range value'))

        # the set now contents individual values
        n3 = set(n3)

        # review range values must have only one hypen second value must be greater than first
        for i in ranges:
            try:
                if item_validator(i[0]) >= item_validator(i[1]):
                    raise ValidationError(_(f'{i[1]} must be greater than {i[0]}'))
            except ValueError:
                raise ValidationError(_(f"{'-'.join(i)} is not a valid range value"))

            for x in range(int(i[0]), int(i[1]) + 1):
                n3.add(x)
        
        # the set is complete
        return n3

    def _minute(self):
        return any([self.minute == '*', self.current_time.minute in self._expanded(self.minute, 'minute')])

    def _hour(self):
        return any([self.hour == '*', self.current_time.hour in self._expanded(self.hour, 'hour')])

    def _day_of_month(self):
        return any([self.day_of_month == '*',
                    self.current_time.day in self._expanded(self.day_of_month, 'day_of_month')])

    def _month(self):
        return any([self.month == '*', self.current_time.month in self._expanded(self.month, 'month')])

    def _day_of_week(self):
        return any([self.day_of_week == '*',
                    self.current_time.isoweekday() in self._expanded(self.day_of_week, 'day_of_week')])

    def must_run(self, date):
        """
        This method is used to check if the Job should be executed by time
        :return: Boolean
        """
        self.current_time = date
        return all([self._minute(), self._hour(), self._day_of_month(), self._month(), self._day_of_week()])


# noinspection SpellCheckingInspection
class Task(models.Model):
    """
    Task class contains the code that will be executed in each task of the process
    """
    offset_validator = RegexValidator(
        r'^-?[0-9]{0,3}%$',
        _('wrong value for offset it mus be positive 1% or negative -1% max 3 digits')
    )
    process = models.ForeignKey(Process, on_delete=models.CASCADE, verbose_name=_("process"), related_name='tasks')
    name = models.CharField(_("name"), max_length=20, unique=True)
    description = models.CharField(_("description"), max_length=200, unique=True)
    is_active = models.BooleanField(_("active"), default=True)
    level = models.PositiveIntegerField(_("diagram level"), default=0)
    offset = models.CharField(_("diagram offset"), max_length=5, default='0%', validators=[offset_validator])
    interpreter = models.CharField(_("interpreter"), max_length=50, blank=True, null=True)
    arguments = models.CharField(_("arguments"), max_length=100, blank=True, null=True)
    code = models.FileField(
        _("code file"),
        upload_to='dj_process_tasks/',
        validators=[FileExtensionValidator(allowed_extensions=['py', 'sh', 'pl'])]
    )
    objects = models.Manager()

    def __str__(self):
        return self.name

    class Meta:
        db_table = 'pr_tasks'
        verbose_name = _('task')
        verbose_name_plural = _('tasks')
        ordering = ['-id']
        permissions = (
            ("view_tasks", "Can view tasks"),
            ("manage_tasks", "Can manage tasks"),
        )

    @property
    def file_extension(self):
        name, extension = os.path.splitext(self.code.name)
        return extension.lstrip('.')


# noinspection SpellCheckingInspection
class TaskDependence(models.Model):
    """
    Identifies the relationship between tasks
    """
    parent = models.ForeignKey(Task, on_delete=models.CASCADE, verbose_name=_("parent task"), related_name='childs')
    task = models.ForeignKey(Task, on_delete=models.CASCADE, verbose_name=_("child task"), related_name='parents')
    objects = models.Manager()

    def __str__(self):
        return f'{self.parent.name}->{self.task.name}'

    class Meta:
        db_table = 'pr_task_dependencies'
        verbose_name = _('task dependence')
        verbose_name_plural = _('task dependencies')
        unique_together = ('parent', 'task')
        ordering = ['-id']
        permissions = (
            ("view_tasks_dependencies", "Can view tasks dependencies"),
        )

    def clean(self):
        super().clean()

        def parent_recursive(task, parent):
            if parent == task:
                raise ValidationError(_('cyclic relation detected'))
            for p in parent.parents.all():
                parent_recursive(task, p.parent)

        parent_recursive(self.task, self.parent)

    def save(self, *args, **kwargs):
        self.full_clean()
        return super().save(*args, **kwargs)


# noinspection SpellCheckingInspection
class Job(models.Model):
    """
    A Job is an instance of a Process that has been executed
    """
    initialized = 'initialized'
    finished = 'finished'
    cancelled = 'cancelled'
    error = 'error'
    cancelable = [initialized, error]
    unfinished = [initialized, error]
    status_choices = (
        (initialized, 'initialized'),
        (finished, 'finished'),
        (cancelled, 'cancelled'),
        (error, 'error'),
    )
    process = models.ForeignKey(
        Process,
        on_delete=models.CASCADE,
        verbose_name=_("process"),
        related_name='jobs'
    )
    status = models.CharField(_("status"), db_index=True, max_length=20, choices=status_choices, default=initialized)
    dt_start = models.DateTimeField(_("start date"), blank=True, null=True, auto_now_add=True)
    dt_end = models.DateTimeField(_("end date"), blank=True, null=True)
    observations = models.CharField(_("observations"), max_length=500, blank=True, null=True)
    objects = models.Manager()

    def __str__(self):
        return f'{self.process.name} [{self.dt_start}][{self.status}]'

    class Meta:
        db_table = 'pr_jobs'
        verbose_name = _('job')
        verbose_name_plural = _('jobs')
        ordering = ['-id']
        permissions = (
            ("view_jobs", "Can view job executions"),
            ("run_jobs", "Can run job"),
            ("cancel_jobs", "Can cancel job"),
        )

    @classmethod
    def create(cls, process, *args, **kwargs):
        """
        this method creates a job-process instance and all its tasks, however if the job its created as finished it will
        not create any task, example the time has come for start a job but a previous instance is still running and the
        process its configured for not overlap the new instance will be created as finished
        """
        job = cls(process=process, *args, **kwargs)
        job.save()
        ret_tasks = []
        if job.status != 'finished':
            tasks = Task.objects.filter(is_active=True, process=process)
            ret_tasks = [JobTask.create(job, t) for t in tasks]
        return job, ret_tasks

    def cancel(self):
        try:
            if self.status not in Job.cancelable:
                raise ProcessException('job status is not initialized')
            with transaction.atomic():
                self.status = Job.cancelled
                # noinspection PyUnresolvedReferences
                self.tasks.filter(status__in=JobTask.can_cancel).update(status=JobTask.cancelled)
                self.save()
        except DatabaseError as e:
            raise ProcessException(e)


# noinspection SpellCheckingInspection
class JobTask(models.Model):
    """
    A JobTask its an instance of a Task that has been executed ant its related to the Job Instance
    """
    initialized = 'initialized'
    awaiting = 'awaiting'
    reopened = 'reopened'
    retry = 'retry'
    finished = 'finished'
    cancelled = 'cancelled'
    forced = 'forced'
    error = 'error'

    can_force = [error]
    can_retry = [error]
    can_reopen = [finished, cancelled, forced]
    can_cancel = [awaiting, error]

    run_status = [awaiting, reopened, retry]
    ok_status = [finished, cancelled, forced]

    trunc_end_dt = [initialized, awaiting, reopened, retry]
    set_end_dt = [finished, cancelled, forced, error]

    status_choices = (
        (awaiting, 'awaiting'),
        (initialized, 'initialized'),
        (finished, 'finished'),
        (cancelled, 'cancelled'),
        (reopened, 'reopened'),
        (forced, 'forced'),
        (retry, 'retry'),
        (error, 'error'),
    )

    # action, requirements, status
    management = (
        ('Force', can_force, forced),
        ('Retry', can_retry, retry),
        ('Reopen', can_reopen, reopened),
        ('Cancel', can_cancel, cancelled),
    )
    job = models.ForeignKey(
        Job,
        on_delete=models.CASCADE,
        verbose_name=_("job"),
        related_name='tasks'
    )
    task = models.ForeignKey(
        Task,
        on_delete=models.CASCADE,
        verbose_name=_("task"),
        related_name='logs'
    )
    status = models.CharField(_("status"), db_index=True, max_length=20, choices=status_choices, default=awaiting)
    dt_created = models.DateTimeField(_("created date"), blank=True, null=True, auto_now_add=True)
    dt_start = models.DateTimeField(_("start date"), blank=True, null=True)
    dt_end = models.DateTimeField(_("end date"), blank=True, null=True)
    observations = models.TextField(_("observations"), blank=True, null=True)
    objects = models.Manager()

    def __str__(self):
        return f'{self.job}[{self.task.name}][{self.status}]'

    class Meta:
        db_table = 'pr_job_tasks'
        verbose_name = _('task instance')
        verbose_name_plural = _('task instances')
        ordering = ['-id']
        permissions = (
            ("view_job_tasks", "Can view tasks executions"),
            ("manage_job_tasks", "Can manage tasks executions"),
        )

    @property
    def title(self):
        return f'{self.status}<br>start: {self.dt_start}<br> end: {self.dt_end}'

    @property
    def info(self):
        if self.status == JobTask.error:
            return f'{self.observations}'
        elif self.status == JobTask.awaiting:
            parents = ', '.join([p.task.name for p in self.get_parents() if p.status != 'finished'])
            return f'waiting for parent(s) {parents}'
        else:
            return self.task.description

    @property
    def ready_to_run(self):
        return all([p.status in JobTask.ok_status for p in self.get_parents()])

    @classmethod
    def create(cls, job, task):
        task = cls(job=job, task=task)
        task.save()
        return task

    def get_childs(self):
        return JobTask.objects.filter(
            job=self.job,
            task__in=TaskDependence.objects.filter(parent=self.task).values('task')
        )

    def get_parents(self):
        return JobTask.objects.filter(
            job=self.job,
            task__in=TaskDependence.objects.filter(task=self.task).values('parent')
        )

    def reopen(self, main=None):
        if main and self.status not in JobTask.can_reopen:
            raise ProcessException(_(f"can't reopen current status not valid"))

        for child in self.get_childs():
            child.reopen()

        self.set_status(JobTask.reopened if main else JobTask.awaiting)

    def set_status(self, status):
        # check available status
        if status not in [i[0] for i in JobTask.status_choices]:
            raise ProcessException(_(f'status requested {status} not in status choices'))
        
        # check if new status is available for current status
        if status == JobTask.cancelled and self.status not in JobTask.can_cancel:
            raise ProcessException(_(f"can't cancel current status not valid"))
        elif status == JobTask.retry and self.status not in JobTask.can_retry:
            raise ProcessException(_(f"can't retry current status not valid"))
        elif status == JobTask.forced and self.status not in JobTask.can_force:
            raise ProcessException(_(f"can't retry current status not valid"))

        # if it will run then trunc observations
        if status in JobTask.run_status:
            self.observations = ''
            if self.job.status in [Job.error, Job.finished]:
                self.job.status = Job.initialized
                self.job.save()

        # DT_START
        if status == JobTask.initialized:
            self.dt_start = timezone.now()
            self.observations = ''

        # DT_END trunc or set
        if status in JobTask.trunc_end_dt:
            self.dt_end = None
        elif status in JobTask.set_end_dt:
            self.dt_end = timezone.now()

        self.status = status
        self.save()



import logging
from django.core.exceptions import ValidationError
from django.contrib import messages
from django.db import transaction
from django.http import HttpResponseRedirect
from django.shortcuts import redirect, get_object_or_404
from django.utils.translation import gettext_lazy as _

from process.conf import get_conf
from .generic_views import (
    ProcessGenericCreateView,
    ProcessGenericListView,
    ProcessGenericUpdateView,
    ProcessGenericDeleteView
)

from process.models import Task, TaskDependence

logger = logging.getLogger('django-process')


class TaskListView(ProcessGenericListView):
    model = Task
    title = 'Tasks'
    filters = get_conf('views__task__list__url_allow_filters')
    permissions = get_conf('views__task__list__permissions')


class TaskDeleteView(ProcessGenericDeleteView):
    model = Task
    success_url = get_conf('views__task__delete__success_url')
    success_message = get_conf('views__task__delete__success_message')
    permissions = get_conf('views__task__delete__permissions')


class TaskCreateView(ProcessGenericCreateView):
    model = Task
    success_url = get_conf('views__task__create__success_url')
    success_message = get_conf('views__task__create__success_message')
    permissions = get_conf('views__task__create__permissions')
    redirect_to_edit = get_conf('views__task__create__redirect_to_edit')

    # noinspection PyTypeChecker
    def post(self, request, *args, **kwargs):
        self.object = None
        form_class = self.get_form_class()
        form = form_class(request.POST, request.FILES)
        if form.is_valid():
            obj = form.save()
            messages.success(request, self.success_message)
            if not self.redirect_to_edit:
                return HttpResponseRedirect(self.get_success_url())
            else:
                return redirect('process-tasks-update', pk=obj.id)

        return self.render_to_response(self.get_context_data(form=form))


class TaskUpdateView(ProcessGenericUpdateView):
    model = Task
    template_name = get_conf('views__templates__task_edit')
    success_url = get_conf('views__task__update__success_url')
    success_message = get_conf('views__task__update__success_message')
    permissions = get_conf('views__task__update__permissions')
    fields = [
        'name',
        'description',
        'is_active',
        'level',
        'offset',
        'interpreter',
        'arguments',
        'code',
    ]

    # noinspection PyUnresolvedReferences
    def get_context_data(self, **kwargs):
        if 'form' not in kwargs:
            kwargs['form'] = self.get_form()

        if 'parents' not in kwargs:
            kwargs['parents'] = self.object.parents.all().\
                extra(select={"badge": "'primary'", 'is_new': "''"}).\
                values('parent__id', 'parent__name', 'badge', 'is_new')

        exclude_ids = [i['parent__id'] for i in kwargs['parents']]
        exclude_ids += [self.object.id]
        if 'new_parents' in kwargs:
            exclude_ids += [i['parent__id'] for i in kwargs['new_parents']]

        # we get the tasks which can be parent
        kwargs['parent_options'] = self.object.process.tasks.all(). \
            exclude(id__in=exclude_ids). \
            values('id', 'name')

        return super().get_context_data(**kwargs)

    # noinspection PyUnresolvedReferences
    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        form = self.get_form()

        parents = self.object.parents.filter(parent__id__in=request.POST.getlist('parents')).\
            extra(select={"badge": "'primary'", 'is_new': "''"}).\
            values('parent__id', 'parent__name', 'badge', 'is_new')

        new_parents = Task.objects.filter(id__in=request.POST.getlist('new-parents')). \
            extra(select={'parent__id': 'id', 'parent__name': 'name', 'badge': "'info'", 'is_new': "'new-'"}). \
            values('parent__id', 'parent__name', 'badge', 'is_new')

        response = self.render_to_response(self.get_context_data(form=form, parents=parents, new_parents=new_parents))
        if not form.is_valid():
            return response

        try:
            with transaction.atomic():
                self.object = form.save()

                # get or create parent relation
                for new_parent in new_parents:
                    new_parent['badge'] = 'danger'
                    parent_task = get_object_or_404(Task, id=new_parent['parent__id'])
                    __, __ = TaskDependence.objects.get_or_create(task=self.object, parent=parent_task)
                    new_parent['badge'] = 'primary'

                for current in self.object.parents.all():
                    if current.parent.id not in [i['parent__id'] for i in parents] + \
                           [i['parent__id'] for i in new_parents]:
                        current.delete()
            messages.success(request, self.success_message)
            return redirect('process-tasks-update', pk=self.object.id)
        except ValidationError as e:
            for msg in e.messages:
                messages.error(request, _(f'{msg}'))
            return response
        except Exception as e:
            messages.error(request, _(f'{e}'))
            return response



import os
import json
import logging
import importlib
from django.conf import settings
from django.db.models import ObjectDoesNotExist
from django.utils import timezone

from process.models import Process, Job, JobTask
from ._task import TaskThreaded

logger = logging.getLogger('django-process')


def configure_env():
    # noinspection PyUnresolvedReferences
    module = importlib.util.find_spec('process')
    mod_location = os.path.dirname(module.origin)
    env_file = os.path.join(mod_location, 'env_conf.json')
    environment = {
        'project_path': settings.BASE_DIR,
        'project_settings': os.environ.get('DJANGO_SETTINGS_MODULE')
    }
    try:
        with open(env_file, 'w') as f:
            json.dump(environment, f)
    except Exception as e:
        logger.exception(f'django-process environment could not be configured due to =>\n{e}')


def run_jobs(date):
    """
    Start Job and it's tasks
    """
    for pr in Process.objects.filter(is_active=True):
        must_run = pr.must_run(date)
        logger.debug(f'process {pr} must run {must_run}')
        if must_run:
            try:
                last_job = Job.objects.filter(process=pr).latest('id')
            except ObjectDoesNotExist:
                # job for the process have never run before set an empty job instance
                # and set status to finished otherwise the overlap validation will avoid
                # a new instance to start
                last_job = Job()
                last_job.status = Job.finished

            if not pr.run_if_err and last_job.status == Job.error:
                logger.error(f'job {pr} will not run because previous job status its error')
                continue

            if not pr.run_overlap and last_job.status == Job.initialized:
                logger.error(f'job {pr} will not run because previous job has not finished '
                             f'and this job does not allow overlap')
                continue

            __, __ = Job.create(pr)


def run_awaiting_tasks():
    """
    Run all pending tasks if their status is pending and their parent tasks are finished
    """
    for task in JobTask.objects.filter(status__in=JobTask.run_status):
        if task.ready_to_run:
            # mark task instance as initialized
            task.set_status(JobTask.initialized)
            TaskThreaded(task).start()


def finish_jobs():
    """
    select all initialized(unfinished) jobs if all it's tasks are finished then finish it self
    """
    def custom_all(plist):
        """
        all() builtin function will return True if an empty list is passed
        :param plist:
        """
        if not plist:
            return False
        return all(plist)
    jobs = Job.objects.filter(status__in=Job.unfinished).prefetch_related()
    for job in jobs:
        if custom_all([i.status in JobTask.ok_status for i in job.tasks.all()]):
            job.status = Job.finished
            job.dt_end = timezone.now()
            job.save()



__title__ = 'django-process'
__description__ = 'An app for create process workflows and schedule tasks on django'
__url__ = 'https://github.com/Jesrat/django-process'
__version__ = '4.5.5'
__author__ = 'Josue Gomez'
__author_email__ = 'jgomez@jesrat.com'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2019 Josue Gomez'



import process as pkg
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    readme = fh.read()

setup(
    name=pkg.__title__,
    version=pkg.__version__,
    author=pkg.__author__,
    author_email=pkg.__author_email__,
    description=pkg.__description__,
    long_description=readme,
    long_description_content_type="text/markdown",
    url=pkg.__url__,
    classifiers=[
        "Programming Language :: Python :: 3",
        'License :: OSI Approved :: Apache Software License',
        'Framework :: Django',
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=['django']
)



import logging
from django.shortcuts import render
from django.views import View

from process.conf import get_conf
from .generic_views import ProcessSecurity


from process.models import Job

logger = logging.getLogger('django-process')


class DiagramView(ProcessSecurity, View):
    model = None
    template = get_conf('views__templates__object_diagram')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.permissions = ['view_jobs'] if self.model == Job else ['view_processes']

    # noinspection PyUnusedLocal
    def get(self, request, pk, *args, **kwargs):
        obj = self.model.objects.get(id=pk)
        return render(request, self.template, {'object': obj})



from django.contrib import admin
from .models import Process, Task, TaskDependence, Job, JobTask


# Register your models here.
@admin.register(Process)
class ProcessAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'is_active', 'minute', 'hour', 'day_of_month', 'month', 'day_of_week')


@admin.register(Task)
class TaskAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'is_active', 'process')


@admin.register(TaskDependence)
class TaskDependenceAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'task', 'parent')


@admin.register(Job)
class JobAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'dt_start', 'dt_end', 'observations')


@admin.register(JobTask)
class JobTaskAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'dt_start', 'dt_end', 'observations')



import logging
from django import template
from django.utils.safestring import mark_safe

from process.conf import get_conf
from process.models import JobTask

logger = logging.getLogger('django-process')

register = template.Library()


@register.filter(name='get_image', is_safe=True)
def get_image(extension):
    path = get_conf(f'views__templates__extension_images__{extension}')
    return path


@register.simple_tag(takes_context=True)
def keep_filters(context, **kwargs):
    d = context['request'].GET.copy()
    for k, v in kwargs.items():
        d[k] = v
    for k in [k for k, v in d.items() if not v]:
        del d[k]
    return d.urlencode()


@register.filter(name='parents_as_badges', is_safe=True)
def parents_as_badges(parents):
    parent_badge_html = """
        <div class="btn btn-{badge} d-flex justify-content-between mb-1" style="cursor: default;">
            <input type="hidden" name="{is_new}parents" value="{id}">
            <span>{name}</span>
            <span class="badge badge-light" style="cursor: pointer;" onclick="removeParentTask(this);">X</span>
        </div>
        """
    response = ''
    for parent in parents:
        response += parent_badge_html.format(
            badge=parent['badge'],
            id=parent['parent__id'],
            name=parent['parent__name'],
            is_new=parent['is_new']
        )
    return mark_safe(response)


# noinspection SpellCheckingInspection
@register.filter(name='available_actions', is_safe=True)
def available_actions(task):
    html = """
    <div class="nav-item dropdown">
        <a class="nav-link dropdown-toggle" href="#" id="object-task-actions" role="button"
           data-toggle="dropdown" aria-haspopup="true" aria-expanded="false">Actions     </a>
        <div class="dropdown-menu" aria-labelledby="object-task-actions">
            {forms}
        </div>
    </div>
    """
    form = """    
    <form method="POST" class="m-0">
        <input type="hidden" name="task" value="{task}">
        <input type="hidden" name="action" value="{status}">
        <button type="submit" class="dropdown-item btn">{action}</button>
    </form>
    """
    actions = []

    for opt_name, opt_list, status in JobTask.management:
        if task.status in opt_list:
            action = form.\
                replace('{task}', str(task.id)).\
                replace('{action}', opt_name).\
                replace('{status}', status)
            actions.append(action)

    response = ''
    if actions:
        response = html.format(forms="\n".join(actions))
    return mark_safe(response)


# noinspection SpellCheckingInspection
@register.filter(name='textarea', is_safe=True)
def textarea(content):
    if not content:
        return content

    rows = len(content.split('\n'))
    return mark_safe(
        '<textarea readonly cols="90" rows="{rows}" style="border-color: #0000000a;">{content}</textarea>'.format(
            **{'content': content, 'rows': 20 if rows > 20 else rows}
        )
    )
